#!/usr/bin/python

import pycurl
import xmltodict
import logging
import platform
from ardi.api import core, subscription, service
from StringIO import StringIO
from subprocess import call
import subprocess
import os
import traceback
import threading
import time
import argparse
import importlib
import requests
import signal

def CatchSignal(a,b):
    signal.signal(signal.SIGINT,signal.SIG_DFL)
    signal.signal(signal.SIGTERM,signal.SIG_DFL)
    note.Stop()

class channel:
    def __init__(self):
        self.ardi = ""
        self.field = ""

class recorder(service.ARDIService):
    def __init__(self):
        super(recorder,self).__init__()
        self.channels = []
        self.image = {}
        self.methods = {}
        self.queue = []
        self.sync = True
        self.queuelock = threading.Lock()
        self.addon = "recorder"
        self.description = "Records data to databases."
        self.writethread = None

    def GetConfiguration(self,server):
        super(recorder,self).GetConfiguration(server)
        try:
             response = StringIO()
            
             fullurl = "http://" + server + "/recorder/api/getchannels?trusted=true"
             #print "Requesting " + fullurl
             c = pycurl.Curl()
             c.setopt(c.URL,fullurl)
             c.setopt(c.WRITEFUNCTION,response.write)
             c.setopt(c.TIMEOUT,5)
             c.setopt(c.CONNECTTIMEOUT,5)
             c.perform()
             ret = response.getvalue()
             c.close()
             return ret
        except:
            return False

    def ProcessConfiguration(self,xml):
        """Imports the configuration XML file

        This function loads the list of data points within the XML text provided. The function is usually called from LoadConfig

        \param xml: The configuration data in XML format"""
        super(recorder,self).ProcessConfiguration(xml)
        result = xmltodict.parse(xml)
        #print xml
                        
        #Core Parameters
        #print str(result)
        self.sync = False
        self.sampletime = 0

        servertype = result['channels']['server']['@driver']
        serveraddr = result['channels']['server']['@address']
        if '@samplerate' in result['channels']['server']:
            if result['channels']['server']['@samplerate'] == 0:
                self.sync = False
            else:
                self.sync = True
                self.sampletime = int(result['channels']['server']['@samplerate'])
            

        print "Bringing up writer for " + servertype + " @ " + serveraddr

        themodule = importlib.import_module(servertype)
        self.writer = themodule.dbwriter(serveraddr)
        self.writer.Connect()
        
        try:
            if '@name' in result['channels']['channel']:
                pass
        except:
            logging.error('The server currently has no recorded data')
            return

        toprocess = result['channels']['channel']
        if '@ardiname' in toprocess:
            toprocess = [toprocess]

        #print str(toprocess)
        
        for itm in toprocess:
            #print "Loaded Channel: " + str(itm['@ardiname'])
            vv = channel()
            vv.ardi = str(itm['@ardiname'])
            vv.field = str(itm['@address'])
            self.channels.append(vv)
            pass
                
        self.writer.Prepare(self.channels)

        self.logger.info("Sample Time: " + str(self.sampletime))

        #Launch the thread to actually perform the writing...
        if self.sampletime != 0:
            self.writethread = threading.Thread(target=self.WriteThread)
            self.writethread.daemon = True
            self.writethread.start()
        else:
            self.writethread = None

    def WriteThread(self):
        self.logger.info("Write Thread Started")
        
        oldwritethread = self.writethread
        while True:
            if self.running == False:
                return
            if self.writethread != oldwritethread:
                return
            
            self.writer.WriteImage(self.image)
            time.sleep(self.sampletime)

        self.logger.info("Write Thread Closed")

    def ProcessNewValues(self,content,params):
        if self.sync == True:
            for v in self.channels:
                if v.ardi in content:
                    self.image[v.field] = content[v.ardi]
            #self.writer.WriteImage(self.image)
        else:
            updates = {}
            for v in self.channels:
                if v.ardi in content:
                    updates[v.field] = content[v.ardi]
                    
            self.writer.WriteImage(updates)

    def Stop(self):
        super(recorder,self).Stop()
        self.writer = None
        

note = recorder()
note.signalhandler = CatchSignal
note.Start()
note.BlockTillEnd()