#!/usr/bin/python

import pycurl
import xmltodict
import logging
import platform
from ardi.api import core, subscription
from StringIO import StringIO
from subprocess import call
import subprocess
import os
import traceback
import threading
import time
import argparse
import importlib
import requests

if platform.system() != "Windows":
    from daemon import runner

def processqueue():
    while True:
        note.ProcessQueue();
        time.sleep(1)
    
def LatestUpdate(code, params):
    #print "New Data " + str(code) + " : " + str(params)
    params.ProcessNewValues(code)
    pass

class channel:
    def __init__(self):
        self.ardi = ""
        self.field = ""

class recorder:
    def __init__(self):
        self.channels = []
        self.image = {}
        self.methods = {}
        self.queue = []
        self.sync = True
        self.queuelock = threading.Lock()
        self.lastsignature = ""
        self.running = True

    def initlogger(self):

        """Initialises the historian logger"""
        logging.basicConfig(level=logging.INFO)
        ##The logger used to record information about the driver process as a whole
        self.logger = logging.getLogger(__name__)
        
        fname = "/var/log/ardi/recorder-" + self.instancename + ".log"
        if (platform.system() == 'Windows'):
            fname = "c:\\windows\\temp\\ardi-recorder-" + self.instancename + ".log"

        loghandler = logging.FileHandler(fname)
        loghandler.setLevel(logging.INFO)

        # create a logging format
        
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        loghandler.setFormatter(formatter)
        
        # add the handlers to the logger
        
        self.logger.addHandler(loghandler)

        logging.getLogger("urllib3").setLevel(logging.WARNING)
        logging.getLogger("requests").setLevel(logging.WARNING)

        self.logger.info('----NOTIFIER RESTART----')
        self.logger.info('Starting Notifier Process')
        
        #redir = iologger(self.logger,logging.ERROR)
        #sys.stderr = redir

    def LoadFromServer(self):
        self.initlogger()
        while True:
            try:
                 servername = self.server
                 response = StringIO()

                 if self.lastsignature == "":
                     fullurl = "http://" + servername + "/recorder/api/version?trusted=true"
                     #print "Requesting " + fullurl
                     c = pycurl.Curl()
                     c.setopt(c.URL,fullurl)
                     c.setopt(c.WRITEFUNCTION,response.write)
                     c.setopt(c.TIMEOUT,5)
                     c.setopt(c.CONNECTTIMEOUT,5)
                     c.perform()
                     ret = response.getvalue()
                     c.close()

                     self.lastsignature = ret

                 response = StringIO()
                
                 fullurl = "http://" + servername + "/recorder/api/getchannels?trusted=true"
                 #print "Requesting " + fullurl
                 c = pycurl.Curl()
                 c.setopt(c.URL,fullurl)
                 c.setopt(c.WRITEFUNCTION,response.write)
                 c.setopt(c.TIMEOUT,5)
                 c.setopt(c.CONNECTTIMEOUT,5)
                 c.perform()
                 ret = response.getvalue()
                 c.close()
                 break
            except:
                 time.sleep(10)
                 pass

        #print "Query Complete - '" + ret + "'"
        #print "Query Finished: '" + ret + "'"
        self.parseconfig(ret)

    def parseconfig(self,xml):
        """Imports the configuration XML file

        This function loads the list of data points within the XML text provided. The function is usually called from LoadConfig

        \param xml: The configuration data in XML format"""
        result = xmltodict.parse(xml)
                        
        #Core Parameters
        #print str(result)

        servertype = result['channels']['server']['@driver']
        serveraddr = result['channels']['server']['@address']
        if '@sync' in result['channels']['server']:
            if result['channels']['server']['@sync'] == 0:
                self.sync = False

        print "Bringing up writer for " + servertype + " @ " + serveraddr

        themodule = importlib.import_module(servertype)
        self.writer = themodule.dbwriter(serveraddr)
        self.writer.Connect()
        
        try:
            if '@name' in result['channels']['channel']:
                pass
        except:
            logging.error('The server currently has no recorded data')
            return

        toprocess = result['channels']['channel']
        if '@ardiname' in toprocess:
            toprocess = [toprocess]
        
        for itm in result['channels']['channel']:
            #print "Loaded Channel: " + str(itm['@ardiname'])
            vv = channel()
            vv.ardi = str(itm['@ardiname'])
            vv.field = str(itm['@address'])
            self.channels.append(vv)
            pass
                

        self.writer.Prepare(self.channels)

    def SetServer(self,ServerName):
        self.server = ServerName

    def ProcessNewValues(self,content):
        if self.sync == True:
            for v in self.channels:
                if v.ardi in content:
                    self.image[v.field] = content[v.ardi]

            self.writer.WriteImage(self.image)
        else:
            updates = {}
            for v in self.channels:
                if v.ardi in content:
                    updates[v.field] = content[v.ardi]
                    
            self.writer.WriteImage(updates)

    def StartWatching(self):
        Cr = core.ARDICore(self.server)
        Cr.Connect()

        self.Sub = subscription.ARDISubscription(Cr)
        for no in self.channels:
            print '  Listening To ' + no.ardi
            self.Sub.AddCode(no.ardi)

        self.Sub.SetCallback(LatestUpdate,self)
        try:
            self.Sub.Connect()
        except:
            traceback.print_exc()

    def ChangeMonitorThread(self):
        self.runthread = threading.Thread(target=self.CheckForChanges)
        self.runthread.daemon = True
        self.runthread.start()

    def CheckForChanges(self):
        while True:
            time.sleep(20)
            print "Checking server for updates..."

            response = StringIO()
            try:
                fullurl = "http://" + self.server + "/recorder/api/version?trusted=true"
                #print "Requesting " + fullurl
                c = pycurl.Curl()
                c.setopt(c.URL,fullurl)
                c.setopt(c.WRITEFUNCTION,response.write)
                c.setopt(c.TIMEOUT,5)
                c.setopt(c.CONNECTTIMEOUT,5)
                c.perform()
                ret = response.getvalue()
                c.close()

                if ret != self.lastsignature:
                    print "Updating with new server data..."
                    self.lastsignature = ret
                    self.Sub.Clear()
                    self.LoadFromServer()
                    for no in self.channels:
                        print '  Listening To ' + no.ardi
                        self.Sub.AddCode(no.ardi)
                    
                    self.Sub.Subscribe()
            except KeyboardInterrupt:
                self.Sub.Disconnect()
                break
            except:
                pass

#    def RunQueueThread(self):
#        self.runthread = threading.Thread(target=processqueue)
#        self.runthread.start()
#        pass


parser = argparse.ArgumentParser(description='Record data from the live ARDI stream to a suitable database / historian.')
if platform.system() != "Windows":
    parser.add_argument('service', help='The instruction you would like to give the service')
parser.add_argument('server', help='The ARDI server/site URL to use')

args = parser.parse_args()

note = recorder()
note.SetServer(args.server)
note.instancename = "default"
ps = args.server.rfind('/')
note.instancename = args.server[ps+1:]

#print "Instance: " + note.instancename

if platform.system() == 'Windows':
    #Daemonise this...
    note.LoadFromServer()
    note.ChangeMonitorThread()
    note.StartWatching()
    pass
else:
    #Just run - no need to daemonise.
    #print "Trying to daemonise this thing!"
        
    class Daemon:
        def __init__(self,inst="default"):
            self.stdin_path = '/dev/null'
            self.stdout_path = '/var/log/ardi/notifier-' + inst
            self.stderr_path = '/var/log/ardi/notifier-' + inst
            self.pidfile_path =  '/var/run/ardi-notifier-' + inst + '.pid'
            self.pidfile_timeout = 5
            
        def run(self):
            note.LoadFromServer()
            note.ChangeMonitorThread()
            note.StartWatching()
            
    if args.server != "standalone":
        dm = Daemon(note.instancename)
        dm.note = note
        daemon_runner = runner.DaemonRunner(dm)
        daemon_runner.do_action()
    else:
        note.LoadFromServer()
        note.ChangeMonitorThread()
        note.StartWatching()
    pass